/*
 * Copyright (c) 2019-2022, 2025 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Mon Mar 11 2019
 */
#include "acmrsim.h"

#include <sys/time.h>
#include <err.h>
#include <pthread.h>

#include <iostream>
#include <unordered_set>
#include <vector>

#include "gz/math/Rand.hh"
#include "gz/plugin/Register.hh"
#include "gz/sim/Joint.hh"
#include "gz/sim/Link.hh"
#include "gz/sim/Model.hh"
#include "gz/sim/System.hh"
#include "gz/sim/Util.hh"
#include "gz/sim/World.hh"
#include "sdf/Element.hh"

#include "libmrsim.h"

using namespace gz;

class mrsim :
  public sim::System,
  public sim::ISystemConfigure,
  public sim::ISystemPreUpdate,
  public sim::ISystemPostUpdate
{
  struct rotor {
    sim::Joint joint;
    sim::Link child, parent;
    math::Vector3d axis;
    double noise;
    int spin;
  };

  pthread_t reader, updater;
  pthread_mutex_t mrsim_mutex = PTHREAD_MUTEX_INITIALIZER;
  struct mrsim_s *mrsim{nullptr};

  sim::Link base;
  std::vector<rotor> rotors;
  bool explicit_wrench;
  int32_t updatep;

 public:

  /* --- Configure --------------------------------------------------------- */

  void
  Configure(const sim::Entity &entity, const sdf::ElementConstPtr &sdf,
            sim::EntityComponentManager &ecm, sim::EventManager &)
  {
    std::unordered_set<sim::Entity> entities;
    struct mrsim_param_s *params;
    std::string name, pty, s;
    sdf::ElementConstPtr b, r, j;
    uint8_t nrotors = 0;
    sim::Model model;
    sim::World world;
    double noise;
    rotor rotor;

    /* get world */
    world = sim::World(worldEntity(entity, ecm));
    if (!world.Valid(ecm)) { gzerr << "[mrsim] no world?\n"; return; }

    /* get model */
    model = sim::Model(entity);
    if (!model.Valid(ecm)) {
      gzerr << "[mrsim] plugin should be attached to a model entity\n";
      return;
    }

    /* get base link */
    if (!sdf->HasElement("link")) {
      gzerr << "[mrsim] missing base <link> tag\n";
      return;
    }
    name = sdf->Get<std::string>("link");
    entities = sim::entitiesFromScopedName(name, ecm, model.Entity());
    if (entities.empty()) {
      gzerr << "[mrsim] non-existent base <link> '" << name << "'\n";
      return;
    }
    base = sim::Link(*entities.begin());
    if (!base.Valid(ecm)) {
      gzerr << "[mrsim] invalid base <link> '" << name << "'\n";
      return;
    }

    /* enable full state reporting */
    base.EnableVelocityChecks(ecm);
    base.EnableAccelerationChecks(ecm);

    /* get list of controlled rotors */
    if (sdf->HasElement("rotors")) {
      r = sdf->FindElement("rotors");

      r->Get<double>("noise", noise, 0.);

      for(r = r->FindElement("rotor"); r; r = r->GetNextElement("rotor")) {
        /* joint */
        j = r->FindElement("joint");
        if (!j) {
          gzerr << "[mrsim] ignoring <rotor> without <joint> tag\n";
          continue;
        }

        name = j->GetValue()->GetAsString();
        entities = sim::entitiesFromScopedName(name, ecm, model.Entity());
        if (entities.empty()) {
          gzerr << "[mrsim] ignoring non-existent joint '" << name << "'\n";
          continue;
        }
        rotor.joint = sim::Joint(*entities.begin());
        if (!rotor.joint.Valid(ecm)) {
          gzerr << "[mrsim] ignoring invalid joint '" << name << "'\n";
          continue;
        }
        if (rotor.joint.Type(ecm).
            value_or(sdf::JointType::INVALID) != sdf::JointType::REVOLUTE) {
          gzerr << "[mrsim] ignoring joint '" << name
                << "' not of the hinge type\n";
          continue;
        }

        /* axis */
        rotor.axis = rotor.joint.Axis(ecm).value()[0].Xyz();

        /* child link */
        s = rotor.joint.ChildLinkName(ecm).value_or("");
        entities = sim::entitiesFromScopedName(s, ecm, model.Entity());
        if (entities.empty()) {
          gzerr << "[mrsim] unknown child '" << s
                << "' in joint '" << name << "'\n";
          continue;
        }
        rotor.child = sim::Link(*entities.begin());
        if (!rotor.child.Valid(ecm)) {
          gzerr << "[mrsim] child '" << s << "' is not a valid link"
                << " in joint '" << name << "'\n";
          continue;
        }

        /* parent link */
        s = rotor.joint.ParentLinkName(ecm).value_or("");
        entities = sim::entitiesFromScopedName(s, ecm, model.Entity());
        if (entities.empty()) {
          gzerr << "[mrsim] unknown parent '" << s
                << "' in joint '" << name << "'\n";
          continue;
        }
        rotor.parent = sim::Link(*entities.begin());
        if (!rotor.parent.Valid(ecm)) {
          gzerr << "[mrsim] parent '" << s << "' is not a valid link"
                << " in joint '" << name << "'\n";
          continue;
        }

        /* joint spin */
        if (j->HasAttribute("spin")) {
          s = j->Get<std::string>("spin");
          if (s == "cw") rotor.spin = -1;
          else if (s == "ccw") rotor.spin = 1;
          else {
            gzerr << "[mrsim] ignoring invalid spin '" << s
                  << "' (must be cw or ccw)\n";
            rotor.spin = 1;
          }
        } else
          rotor.spin = 1;

        /* axis noise */
        r->Get<double>("noise", rotor.noise, noise);

        /* just for visual effect at startup */
        rotor.joint.ResetPosition(ecm, { math::Rand::DblUniform(0, M_PI) });

        /* push */
        rotors.push_back(rotor);
      }
      nrotors = rotors.size();
    }

    if (sdf->HasElement("allocation")) {
      /* get number of rotors from matrix */
      std::stringstream G(sdf->Get<std::string>("allocation"));
      double d;

      for(nrotors = 0; !G.fail(); nrotors++) G >> d;
      nrotors /= 6;
    }


    /* init mrsim */
    struct timespec now;

    mrsim = mrsim_init(nrotors);
    if (!mrsim) {
      gzerr << "[mrsim] could not initialize libmrsim\n";
      return;
    }

    clock_gettime(CLOCK_REALTIME, &now);
    mrsim_update_time(mrsim, now);

    sdf->Get<std::string>("pty", pty, "/tmp/pty-" + model.Name(ecm));

    if (mrsim_open_pty(mrsim, pty.c_str()))
      gzerr << "[mrsim] cannot create pty " << pty << "\n";

    params = mrsim_params(mrsim);
    params->body.g =
      world.Gravity(ecm).value_or(math::Vector3d(0, 0, 9.81)).Length();
    math::Vector3d B =
      world.MagneticField(ecm).
      value_or(math::Vector3d(23.816, -0.41, -39.829) * 1e-6);
    params->body.B[0] = B.X();
    params->body.B[1] = B.Y();
    params->body.B[2] = B.Z();

    if (sdf->HasElement("rotors")) {
      r = sdf->FindElement("rotors");

      if (rotors.size() > 0) {
        auto I = rotors[0].child.WorldInertial(ecm);
        if (I.has_value())
          params->motor.J = I.value().Moi()(2,2);
      }

      if (r->HasElement("J")) params->motor.J = r->Get<double>("J");
      if (r->HasElement("b")) params->motor.b = r->Get<double>("b");
      if (r->HasElement("K")) params->motor.K = r->Get<double>("K");
      if (r->HasElement("R")) params->motor.R = r->Get<double>("R");
      if (r->HasElement("L")) params->motor.L = r->Get<double>("L");
      if (r->HasElement("V")) params->motor.V = r->Get<double>("V");
      if (r->HasElement("I")) params->motor.I = r->Get<double>("I");
      if (r->HasElement("cf")) params->motor.Cf = r->Get<double>("cf");
      if (r->HasElement("ct")) params->motor.Ct = r->Get<double>("ct");
      if (r->HasElement("vfxy")) params->motor.Vfxy = r->Get<double>("vfxy");
      if (r->HasElement("vfz")) params->motor.Vfz = r->Get<double>("vfz");
      if (r->HasElement("vtxy")) params->motor.Vtxy = r->Get<double>("vtxy");
      if (r->HasElement("vtz")) params->motor.Vtz = r->Get<double>("vtz");
    }

    explicit_wrench = sdf->HasElement("allocation");
    if (explicit_wrench) {
      std::stringstream G(sdf->Get<std::string>("allocation"));
      uint8_t i;

      for(i = 0; i < 6 * params->body.rotors; i++) {
        G >> params->body.G[i];
        if (G.fail()) {
          gzerr << "[mrsim] ignoring malformed 6×" << params->body.rotors
                << " allocation matrix\n";
          explicit_wrench = false;
          break;
        }
      }
      if (explicit_wrench) {
        G >> std::ws;
        if (!G.eof()) {
          gzerr << "[mrsim] ignoring malformed 6×" << params->body.rotors
                << " allocation matrix\n";
          explicit_wrench = false;
        }
      }
    }


    /* create reader and updater thread */
    updatep =
      1000000 * (sdf->HasElement("period") ? sdf->Get<double>("period") : 1e-3);
    pthread_create(&reader, NULL, reader_thread, this);
    pthread_create(&updater, NULL, updater_thread, this);

    warnx("mrsim created %s", pty.c_str());
  }


  /* --- PreUpdate --------------------------------------------------------- */

  void
  PreUpdate(const sim::UpdateInfo &info, sim::EntityComponentManager &ecm)
  {
    std::vector<rotor>::const_iterator r;
    struct mrsimstate_wrench_s wrench;
    struct mrsimstate_motor_s *motor;
    math::Quaterniond q;
    math::Vector3d z;
    double zaxis[3];
    uint8_t i;

    if (info.paused) return;

    pthread_mutex_lock(&mrsim_mutex);
    if (!mrsim) {
      pthread_mutex_unlock(&mrsim_mutex);
      return;
    }

    /* set rotor velocity for visual effect - scale down by 6.28 */
    for(i = 0; i < rotors.size(); i++) {
      motor = mrsim_motor(mrsim, i);

      rotors[i].joint.ResetVelocity(
        ecm, {rotors[i].spin * copysign(motor->w, motor->spin)});
    }

    if (explicit_wrench) {
      /* apply explicit wrench */
      wrench = *mrsim_wrench(mrsim);

      q = base.WorldPose(ecm).value_or(math::Pose3d::Zero).Rot();
      base.AddWorldWrench(
        ecm,
        q * math::Vector3d(
          wrench.thrust[0], wrench.thrust[1], wrench.thrust[2]),
        q * math::Vector3d(
          wrench.torque[0], wrench.torque[1], wrench.torque[2]));
    } else {
      /* apply thrust to rotors and drag to base link */
      for(i = 0, r = rotors.begin(); r != rotors.end(); ++r, i++) {
        z = r->child.WorldPose(ecm).value_or(math::Pose3d::Zero).Rot()
            * r->axis;
        if (r->noise > 0.)
          z = math::Quaterniond(
            math::Rand::DblNormal(0., r->noise),
            math::Rand::DblNormal(0., r->noise),
            math::Rand::DblNormal(0., r->noise)) * z;
        zaxis[0] = z.X(); zaxis[1] = z.Y(); zaxis[2] = z.Z();

        wrench = mrsim_motor_wrench(mrsim, i, zaxis);

        r->child.AddWorldForce(
          ecm, { wrench.thrust[0], wrench.thrust[1], wrench.thrust[2] });
        r->parent.AddWorldWrench(
          ecm, math::Vector3d::Zero,
          rotors[i].spin * math::Vector3d{
            wrench.torque[0], wrench.torque[1], wrench.torque[2]
          });
      }
    }

    pthread_mutex_unlock(&mrsim_mutex);
  }


  /* --- PostUpdate -------------------------------------------------------- */

  void
  PostUpdate(const sim::UpdateInfo &info,
             const sim::EntityComponentManager &ecm)
  {
    math::Pose3d p;
    math::Vector3d v, w, a(0., 0., 0.), dw(0., 0., 0.);
    struct mrsimstate_body_s *state;

    pthread_mutex_lock(&mrsim_mutex);
    if (!mrsim) {
      pthread_mutex_unlock(&mrsim_mutex);
      return;
    }

    /* get state and update mrsim */
    p = base.WorldPose(ecm).value_or(math::Pose3d::Zero);
    v = base.WorldLinearVelocity(ecm).value_or(math::Vector3d::Zero);
    w = base.WorldAngularVelocity(ecm).value_or(math::Vector3d::Zero);
    a = base.WorldLinearAcceleration(ecm).value_or(math::Vector3d::Zero);
    dw = base.WorldAngularAcceleration(ecm).value_or(math::Vector3d::Zero);

    state = mrsim_body(mrsim);
    state->px = p.Pos().X(); state->py = p.Pos().Y(); state->pz = p.Pos().Z();
    state->qw = p.Rot().W();
    state->qx = p.Rot().X(); state->qy = p.Rot().Y(); state->qz = p.Rot().Z();
    state->vx = v.X(); state->vy = v.Y(); state->vz = v.Z();
    state->wx = w.X(); state->wy = w.Y(); state->wz = w.Z();
    state->ax = a.X(); state->ay = a.Y(); state->az = a.Z();
    state->dwx = dw.X(); state->dwy = dw.Y(); state->dwz = dw.Z();

    mrsim_update_imu(mrsim);
    pthread_mutex_unlock(&mrsim_mutex);
  }


  /* --- ~mrsim ------------------------------------------------------------ */

  ~mrsim()
  {
    struct mrsim_s *handle;

    pthread_mutex_lock(&mrsim_mutex);
    handle = mrsim;
    mrsim = NULL;
    pthread_mutex_unlock(&mrsim_mutex);
    if (!handle) return;

    pthread_join(reader, NULL);
    pthread_join(updater, NULL);
    mrsim_fini(handle);
  }


 private:
  /* --- reader_thread ----------------------------------------------------- */

  static void *
  reader_thread(void *data)
  {
    ::mrsim *plugin = (::mrsim *)data;
    int s;

    pthread_mutex_lock(&plugin->mrsim_mutex);
    while(plugin->mrsim) {
      pthread_mutex_unlock(&plugin->mrsim_mutex);

      s = mrsim_wait_msg(plugin->mrsim, 500);

      pthread_mutex_lock(&plugin->mrsim_mutex);
      if (!plugin->mrsim) break;
      if (s == 0) continue;

      if (s < 0 || mrsim_process_msgs(plugin->mrsim)) {
        gzerr << "[mrsim] failed to process message\n";
        break;
      }
    }
    pthread_mutex_unlock(&plugin->mrsim_mutex);

    warnx("mrsim disconnecting");
    return NULL;
  }


  /* --- updater_thread ---------------------------------------------------- */

  static void *
  updater_thread(void *data)
  {
    ::mrsim *plugin = (::mrsim *)data;
    struct timeval dl, tv;
    struct timespec req, now;

    gettimeofday(&dl, NULL);

    pthread_mutex_lock(&plugin->mrsim_mutex);
    while(plugin->mrsim) {
      pthread_mutex_unlock(&plugin->mrsim_mutex);

      /* next deadline - skip missed periods */
      gettimeofday(&tv, NULL);
      do {
        dl.tv_usec += plugin->updatep;
        while (dl.tv_usec >= 1000000) {
          dl.tv_usec -= 1000000;
          dl.tv_sec++;
        }
      } while(dl.tv_sec < tv.tv_sec ||
              (dl.tv_sec == tv.tv_sec && dl.tv_usec < tv.tv_usec));

      /* sleep time */
      if (dl.tv_usec < tv.tv_usec) {
        tv.tv_usec -= 1000000;
        tv.tv_sec++;
      }
      req.tv_sec = dl.tv_sec - tv.tv_sec;
      req.tv_nsec = 1000 * (dl.tv_usec - tv.tv_usec);
      while(nanosleep(&req, &now)) {
        if (errno != EINTR) { gzerr << "[mrsim] nanosleep failed"; break; }
        req = now;
      }

      pthread_mutex_lock(&plugin->mrsim_mutex);
      if (!plugin->mrsim) break;

      /* update simulation */
      clock_gettime(CLOCK_REALTIME, &now);
      mrsim_update_time(plugin->mrsim, now);
      mrsim_update_motors(plugin->mrsim);
      if (plugin->explicit_wrench) mrsim_update_wrench(plugin->mrsim);

      /* send sensor messages */
      if (mrsim_send_msgs(plugin->mrsim))
        gzerr << "[mrsim] failed to send messages\n";
    }
    pthread_mutex_unlock(&plugin->mrsim_mutex);

    return NULL;
  }
};

/* register plugin */
GZ_ADD_PLUGIN(
  mrsim,
  sim::System,
  mrsim::ISystemConfigure,
  mrsim::ISystemPreUpdate,
  mrsim::ISystemPostUpdate)
